@@ -1,13 +1,15 @@
-import os
-
 from sqlobject import col, dberrors
 from sqlobject.compat import PY2
+from sqlobject.converters import registerConverter, StringLikeConverter
 from sqlobject.dbconnection import DBAPI
 
 
 class ErrorMessage(str):
     def __new__(cls, e, append_msg=''):
-        obj = str.__new__(cls, e.args[1] + append_msg)
+        if len(e.args) > 1:
+            obj = str.__new__(cls, e.args[1] + append_msg)
+        else:
+            obj = str.__new__(cls, append_msg)
         try:
             obj.code = int(e.args[0])
         except ValueError:
@@ -72,6 +74,9 @@ class MySQLConnection(DBAPI):
                         oursql.errnos['CR_SERVER_GONE_ERROR']
                     self.CR_SERVER_LOST = oursql.errnos['CR_SERVER_LOST']
                     self.ER_DUP_ENTRY = oursql.errnos['ER_DUP_ENTRY']
+                elif driver == 'mariadb':
+                    import mariadb
+                    self.module = mariadb
                 elif driver == 'pyodbc':
                     import pyodbc
                     self.module = pyodbc
@@ -88,7 +93,7 @@ class MySQLConnection(DBAPI):
                     raise ValueError(
                         'Unknown MySQL driver "%s", '
                         'expected mysqldb, connector, '
-                        'oursql, pymysql, '
+                        'oursql, pymysql, mariadb, '
                         'odbc, pyodbc or pypyodbc' % driver)
             except ImportError:
                 pass
@@ -97,12 +102,14 @@ class MySQLConnection(DBAPI):
         else:
             raise ImportError(
                 'Cannot find a MySQL driver, tried %s' % drivers)
+
         self.host = host
         self.port = port or 3306
         self.db = db
         self.user = user
         self.password = password
         self.kw = {}
+
         for key in ("unix_socket", "init_command",
                     "read_default_file", "read_default_group", "conv"):
             if key in kw:
@@ -127,21 +134,29 @@ class MySQLConnection(DBAPI):
             self.dbEncoding = None
         self.driver = driver
 
+        if driver in ('mariadb', 'odbc', 'pyodbc', 'pypyodbc'):
+            self.CR_SERVER_GONE_ERROR = 2006
+            self.CR_SERVER_LOST = 2013
+            self.ER_DUP_ENTRY = '23000'
+
         if driver in ('odbc', 'pyodbc', 'pypyodbc'):
             self.make_odbc_conn_str(kw.pop('odbcdrv',
                                            'MySQL ODBC 5.3 ANSI Driver'),
                                     db, host, port, user, password
                                     )
-            self.CR_SERVER_GONE_ERROR = 2006
-            self.CR_SERVER_LOST = 2013
-            self.ER_DUP_ENTRY = '23000'
 
-        elif self.driver == 'oursql':
+        elif driver == 'oursql':
             if "use_unicode" not in self.kw:
                 self.kw["use_unicode"] = not PY2
             # oursql doesn't implement ping(True) yet
             self.kw["autoreconnect"] = True
 
+        elif driver == 'mariadb':
+            self.kw.pop("charset", None)
+
+        elif driver == 'connector':
+            registerConverter(bytes, ConnectorBytesConverter)
+
         global mysql_Bin
         if not PY2 and mysql_Bin is None:
             mysql_Bin = self.module.Binary
@@ -150,6 +165,8 @@ class MySQLConnection(DBAPI):
 
         self._server_version = None
         self._can_use_microseconds = None
+        self._can_use_json_funcs = None
+
         DBAPI.__init__(self, **kw)
 
     @classmethod
@@ -179,7 +196,13 @@ class MySQLConnection(DBAPI):
                 conn = self.module.connect(
                     host=self.host, port=self.port, db=self.db,
                     user=self.user, passwd=self.password, **self.kw)
-                if self.driver != 'oursql':
+                if self.driver == 'mariadb':
+                    # Attempt to reconnect.
+                    # This setting is persistent due to ``auto_reconnect``.
+                    # mariadb doesn't implement ping(True)
+                    conn.auto_reconnect = True
+                    conn.ping()
+                elif self.driver != 'oursql':
                     # Attempt to reconnect. This setting is persistent.
                     conn.ping(True)
         except self.module.OperationalError as e:
@@ -229,7 +252,7 @@ class MySQLConnection(DBAPI):
             self.printDebug(conn, query, 'QueryR')
         dbEncoding = self.dbEncoding
         if dbEncoding and not isinstance(query, bytes) and (
-                self.driver in ('mysqldb', 'connector', 'oursql')):
+                self.driver in ('mysqldb', 'connector', 'oursql', 'mariadb')):
             query = query.encode(dbEncoding, 'surrogateescape')
         # When a server connection is lost and a query is attempted, most of
         # the time the query will raise a SERVER_LOST exception, then at the
@@ -263,6 +286,9 @@ class MySQLConnection(DBAPI):
                 msg = ErrorMessage(e)
                 if e.args[0] == self.ER_DUP_ENTRY:
                     raise dberrors.DuplicateEntryError(msg)
+                elif isinstance(e.args[0], str) \
+                        and e.args[0].startswith('Duplicate'):
+                    raise dberrors.DuplicateEntryError(msg)
                 else:
                     raise dberrors.IntegrityError(msg)
             except self.module.InternalError as e:
@@ -345,6 +371,8 @@ class MySQLConnection(DBAPI):
         except dberrors.ProgrammingError as e:
             if e.args[0].code in (1146, '42S02'):  # ER_NO_SUCH_TABLE
                 return False
+            if self.driver == 'mariadb':
+                return False
             raise
 
     def addColumn(self, tableName, column):
@@ -375,6 +403,9 @@ class MySQLConnection(DBAPI):
             # (SQLObject expected '')
             kw['notNone'] = (nullAllowed.upper() != 'YES' and True or False)
 
+            if not PY2 and isinstance(t, bytes):
+                t = t.decode('ascii')
+
             if default and t.startswith('int'):
                 kw['default'] = int(default)
             elif default and t.startswith('float'):
@@ -391,6 +422,8 @@ class MySQLConnection(DBAPI):
         return results
 
     def guessClass(self, t):
+        if not PY2 and isinstance(t, bytes):
+            t = t.decode('ascii')
         if t.startswith('int'):
             return col.IntCol, {}
         elif t.startswith('enum'):
@@ -447,10 +480,10 @@ class MySQLConnection(DBAPI):
             return col.Col, {}
 
     def listTables(self):
-        return [v[0] for v in self.queryAll("SHOW TABLES")]
+        return _decodeBytearrays(self.queryAll("SHOW TABLES"))
 
     def listDatabases(self):
-        return [v[0] for v in self.queryAll("SHOW DATABASES")]
+        return _decodeBytearrays(self.queryAll("SHOW DATABASES"))
 
     def _createOrDropDatabase(self, op="CREATE"):
         self.query('%s DATABASE %s' % (op, self.db))
@@ -482,9 +515,6 @@ class MySQLConnection(DBAPI):
     def can_use_microseconds(self):
         if self._can_use_microseconds is not None:
             return self._can_use_microseconds
-        if os.environ.get('APPVEYOR') or os.environ.get('TRAVIS'):
-            self._can_use_microseconds = False
-            return False
         server_version = self.server_version()
         if server_version is None:
             return None
@@ -495,3 +525,33 @@ class MySQLConnection(DBAPI):
             can_use_microseconds = (server_version >= (5, 6, 4))
         self._can_use_microseconds = can_use_microseconds
         return can_use_microseconds
+
+    def can_use_json_funcs(self):
+        if self._can_use_json_funcs is not None:
+            return self._can_use_json_funcs
+        server_version = self.server_version()
+        if server_version is None:
+            return None
+        server_version, db_tag = server_version
+        if db_tag == "MariaDB":
+            can_use_json_funcs = (server_version >= (10, 2, 7))
+        else:  # MySQL
+            can_use_json_funcs = (server_version >= (5, 7, 0))
+        self._can_use_json_funcs = can_use_json_funcs
+        return can_use_json_funcs
+
+
+def ConnectorBytesConverter(value, db):
+    if not PY2:
+        # For PY2 this converter is called also for SQLite
+        assert db == 'mysql'
+        value = value.decode('latin1')
+    return StringLikeConverter(value, db)
+
+
+def _decodeBytearrays(v_list):
+    if not v_list:
+        return []
+    if not PY2 and isinstance(v_list[0][0], bytearray):
+        return [v[0].decode('ascii') for v in v_list]
+    return [v[0] for v in v_list]
